from laktory._testing import StockPriceSource


def test_eventsource():
    source = StockPriceSource()

    assert source.producer.name == "yahoo-finance"
    assert source.read_as_stream
    assert (
        source.event_root
        == "/Volumes/dev/sources/landing/events/yahoo-finance/stock_price/"
    )
    assert source.fmt == "JSON"
    assert not source.multiline


if __name__ == "__main__":
    test_eventsource()



from laktory.models.datasources import EventDataSource


def test_event_data_source():
    reader = EventDataSource(
        name="stock_price",
        producer={"name": "yahoo_finance"},
    )
    assert (
        reader.event_root
        == "/Volumes/dev/sources/landing/events/yahoo_finance/stock_price/"
    )


if __name__ == "__main__":
    test_event_data_source()



from laktory.models import DataEvent
from laktory.models import Producer
from laktory.models import Pipeline
from laktory.models import Table
from laktory.models import EventDataSource


class StockPriceSource(EventDataSource):
    name: str = "stock_price"
    producer: Producer = Producer(name="yahoo-finance")


class StockPricesPipeline(Pipeline):
    name: str = "pl-stock-prices"
    tables: list[Table] = [
        Table(
            **{
                "name": "brz_stock_prices",
                "timestamp_key": "data.created_at",
                "event_source": StockPriceSource(),
                "zone": "BRONZE",
            }
        ),
        Table(
            **{
                "name": "slv_stock_prices",
                "table_source": {"name": "brz_stock_prices"},
                "zone": "SILVER",
                "columns": [
                    {
                        "name": "created_at",
                        "type": "timestamp",
                        "func_name": "coalesce",
                        "input_cols": ["_created_at"],
                    },
                    {
                        "name": "low",
                        "type": "double",
                        "func_name": "coalesce",
                        "input_cols": ["data.low"],
                    },
                    {
                        "name": "high",
                        "type": "double",
                        "func_name": "coalesce",
                        "input_cols": ["data.high"],
                    },
                ],
            }
        ),
    ]


if __name__ == "__main__":
    # Publish Metadata
    pl = StockPricesPipeline()
    pl.publish_tables_meta()
    print(pl)

    # TODO: push pipeline with Pulumi



import json
from typing import Literal
from typing import Any
from typing import Union

from pydantic import computed_field
from pydantic import model_validator

from laktory.spark import DataFrame
from laktory.spark import df_has_column

from laktory._logger import get_logger
from laktory.models.base import BaseModel
from laktory.models.sql.column import Column
from laktory.models.datasources.tabledatasource import TableDataSource
from laktory.models.datasources.eventdatasource import EventDataSource
from laktory.models.grants.tablegrant import TableGrant

logger = get_logger(__name__)


class Table(BaseModel):
    name: str
    columns: list[Column] = []
    primary_key: Union[str, None] = None
    comment: Union[str, None] = None
    catalog_name: Union[str, None] = None
    schema_name: Union[str, None] = None
    grants: list[TableGrant] = None

    # Data
    data: list[list[Any]] = None

    # Lakehouse
    timestamp_key: Union[str, None] = None
    event_source: Union[EventDataSource, None] = None
    table_source: Union[TableDataSource, None] = None
    zone: Literal["BRONZE", "SILVER", "SILVER_STAR", "GOLD"] = None
    pipeline_name: Union[str, None] = None
    # joins
    # expectations

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    @model_validator(mode="after")
    def assign_table_to_columns(self) -> Any:
        for c in self.columns:
            c.table_name = self.name
            c.catalog_name = self.catalog_name
            c.schema_name = self.schema_name
        return self

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def database_name(self) -> str:
        return self.schema_name

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def column_names(self):
        return [c.name for c in self.columns]

    @property
    def df(self):
        import pandas as pd

        return pd.DataFrame(data=self.data, columns=self.column_names)

    @property
    def source(self):
        if self.event_source is not None and self.event_source.name is not None:
            return self.event_source
        elif self.table_source is not None and self.table_source.name is not None:
            return self.table_source

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #
    #
    # @classmethod
    # def meta_table(cls):
    #     # Build columns
    #     columns = []
    #     for k, t in cls.model_serialized_types().items():
    #         jsonize = False
    #         # if k in ["columns", "event_source", "table_source"]:
    #         if k in ["columns"]:
    #             t = "string"
    #             jsonize = True
    #
    #         elif k in ["data"]:
    #             continue
    #
    #         columns += [
    #             Column(name=k, type=py_to_sql(t, mode="schema"), jsonize=jsonize)
    #         ]
    #
    #     # Set table
    #     return Table(
    #         name="tables",
    #         schema_name="laktory",
    #         columns=columns,
    #     )

    # ----------------------------------------------------------------------- #
    # SQL Methods                                                             #
    # ----------------------------------------------------------------------- #
    #
    # def exists(self):
    #     return self.name in [
    #         c.name
    #         for c in self.workspace_client.tables.list(
    #             catalog_name=self.catalog_name,
    #             schema_name=self.schema_name,
    #         )
    #     ]
    #
    # def create(
    #     self,
    #     if_not_exists: bool = True,
    #     or_replace: bool = False,
    #     insert_data: bool = False,
    #     warehouse_id: str = None,
    # ):
    #     if len(self.columns) == 0:
    #         raise ValueError()
    #
    #     if or_replace:
    #         if_not_exists = False
    #
    #     statement = f"CREATE "
    #     if or_replace:
    #         statement += "OR REPLACE "
    #     statement += "TABLE "
    #     if if_not_exists:
    #         statement += "IF NOT EXISTS "
    #
    #     statement += f"{self.schema_name}.{self.name}"
    #
    #     statement += "\n   ("
    #     for c in self.columns:
    #         t = c.type
    #         if c.jsonize:
    #             t = "string"
    #         statement += f"{c.name} {t},"
    #     statement = statement[:-1]
    #     statement += ")"
    #
    #     logger.info(statement)
    #     r = self.workspace_client.execute_statement_and_wait(
    #         statement, warehouse_id=warehouse_id, catalog_name=self.catalog_name
    #     )
    #
    #     if insert_data:
    #         self.insert()
    #
    #     return r
    #
    # def delete(self):
    #     self.workspace_client.tables.delete(self.full_name)
    #
    # def insert(self, warehouse_id: str = None):
    #     if self.data is None or len(self.data) == 0:
    #         return
    #
    #     statement = f"INSERT INTO {self.full_name} VALUES\n"
    #     for row in self.data:
    #         statement += "   ("
    #         values = [
    #             json.dumps(v) if c.jsonize else v for c, v in zip(self.columns, row)
    #         ]
    #         values = [py_to_sql(v) for v in values]
    #         statement += ", ".join(values)
    #         statement += "),\n"
    #
    #     statement = statement[:-2]
    #
    #     logger.info(statement)
    #     r = self.workspace_client.execute_statement_and_wait(
    #         statement, warehouse_id=warehouse_id, catalog_name=self.catalog_name
    #     )
    #
    #     return r
    #
    # def select(self, limit=10, warehouse_id: str = None, load_json=True):
    #     statement = f"SELECT * from {self.full_name} limit {limit}"
    #
    #     r = self.workspace_client.execute_statement_and_wait(
    #         statement, warehouse_id=warehouse_id, catalog_name=self.catalog_name
    #     )
    #
    #     data = r.result.data_array
    #
    #     if load_json:
    #         for i in range(len(data)):
    #             j = -1
    #             for c, v in zip(self.columns, data[i]):
    #                 j += 1
    #                 if c.jsonize:
    #                     data[i][j] = json.loads(data[i][j])
    #
    #     return data

    # ----------------------------------------------------------------------- #
    # Pipeline Methods                                                        #
    # ----------------------------------------------------------------------- #

    def read_source(self, spark) -> DataFrame:
        return self.source.read(spark)

    def process_bronze(self, df) -> DataFrame:
        import pyspark.sql.functions as F

        logger.info(f"Applying bronze transformations")

        df = df.withColumn("_bronze_at", F.current_timestamp())

        return df

    def process_silver(self, df, table) -> DataFrame:
        import pyspark.sql.functions as F

        logger.info(f"Applying silver transformations")

        columns = []

        # User defined columns
        columns += table.columns

        if table.timestamp_key is not None:
            columns += [
                Column(
                    **{
                        "name": "_tstamp",
                        "type": "timestamp",
                        "func_name": "coalesce",
                        "input_cols": [table.timestamp_key],
                    }
                )
            ]

        # Timestamps
        if df_has_column(df, "bronze_at"):
            columns += [
                Column(
                    **{
                        "name": "_bronze_at",
                        "type": "timestamp",
                        "func_name": "coalesce",
                        "input_cols": ["_bronze_at", "bronze_at"],
                    }
                )
            ]
        columns += [
            Column(
                **{
                    "name": "_silver_at",
                    "type": "timestamp",
                    "func_name": "current_timestamp",
                    "input_cols": [],
                }
            )
        ]

        # Saved existing column names
        cols0 = [v[0] for v in df.dtypes]

        # Build new columns
        logger.info(f"Setting silver columns...")
        new_col_names = []
        for col in columns:
            # Get column definition
            col_name = col.name
            col_type = col.type
            func_name = col.func_name

            #     # TODO: Support customer functions
            #     # if udf_name in udfuncs.keys():
            #     #     f = udfuncs[udf_name]
            #     # else:
            #     #     f = getattr(F, udf_name)

            f = getattr(F, func_name)
            new_col_names += [col_name]

            input_cols = col.input_cols
            kwargs = col.func_kwargs

            #     # Issue when setting strings
            #     # if "format" in kwargs and udf_name == "date_format":
            #     #     kwargs["format"] = kwargs["format"].replace("T", "'T'").replace("''T''", "'T'")

            logger.info(
                f"   {col_name}[{col_type}] as {func_name}({input_cols}, {kwargs})"
            )

            # Cast inputs
            tmp_input_cols = []
            for i, input_col in enumerate(input_cols):
                if func_name == "coalesce" and not df_has_column(df, input_col):
                    # Because the `coalesce` function supports multiple
                    # optional input columns, we can skip the ones not
                    # available in the dataframe
                    logger.warning(f"Column '{input_col}' not available")
                    continue

                tmp_input_cols += [f"__{i}"]
                _icol = tmp_input_cols[-1]
                df = df.withColumn(_icol, F.expr(input_col))
                if input_col.startswith("data.") or func_name == "coalesce":
                    input_type = dict(df.dtypes)[_icol]
                    if input_type in ["double"]:
                        # Some bronze NaN data will be converted to 0 if cast to int
                        sdf = df.withColumn(
                            _icol, F.when(F.isnan(_icol), None).otherwise(F.col(_icol))
                        )
                    if col_type not in ["_any"] and func_name not in [
                        "to_safe_timestamp"
                    ]:
                        df = df.withColumn(_icol, F.col(_icol).cast(col_type))

            if len(input_cols) > 0 and len(tmp_input_cols) == 0:
                raise ValueError(
                    f"None of the inputs columns ({input_cols}) for {col_name} have been found"
                )

            # Set output
            df = df.withColumn(
                col_name,
                f(
                    *tmp_input_cols,
                    **kwargs,
                ),
            )

            # Drop temp inputs
            if len(tmp_input_cols) > 0:
                df = df.drop(*tmp_input_cols)

            # Remove from drop list
            if col_name in cols0:
                cols0.remove(col_name)

        # Drop previous columns
        logger.info(f"Dropping bronze columns...")
        df = df.select(new_col_names)

        # ------------------------------------------------------------------- #
        # Setting Watermark                                                   #
        # ------------------------------------------------------------------- #

        # TODO:
        # if watermark is not None:
        #     sdf = sdf.withWatermark(watermark["column"], watermark["threshold"])

        # ------------------------------------------------------------------- #
        # Drop duplicates                                                     #
        # ------------------------------------------------------------------- #

        pk = table.primary_key
        if pk:
            logger.info(f"Removing duplicates with {pk}")
            df = df.dropDuplicates([pk])

        return df

    def process_silver_star(self, df) -> DataFrame:
        return df



import pytest
from datetime import datetime
from pydantic import ValidationError
import pandas as pd

from laktory.models import Catalog
from laktory.models import Schema
from laktory.models import Column
from laktory.models import Table
from laktory.models import EventDataSource


table = Table(
    name="googl",
    columns=[
        {
            "name": "open",
            "type": "double",
        },
        {
            "name": "close",
            "type": "double",
        },
    ],
    data=[[1, 2], [3, 4], [5, 6]],
    zone="SILVER",
    catalog_name="lakehouse",
    schema_name="markets",
    event_source=EventDataSource(
        name="stock_price",
    ),
)


def test_data():
    assert table.df.equals(
        pd.DataFrame(
            {
                "open": [1, 3, 5],
                "close": [2, 4, 6],
            }
        )
    )


def test_model():
    assert table.columns == [
        Column(
            name="open",
            type="double",
            catalog_name="lakehouse",
            schema_name="markets",
            table_name="googl",
        ),
        Column(
            name="close",
            type="double",
            catalog_name="lakehouse",
            schema_name="markets",
            table_name="googl",
        ),
    ]
    assert table.catalog_name == "lakehouse"
    assert table.schema_name == "markets"
    assert table.parent_full_name == "lakehouse.markets"
    assert table.full_name == "lakehouse.markets.googl"
    assert table.zone == "SILVER"
    assert table.source.name == "stock_price"

    # Invalid zone
    with pytest.raises(ValidationError):
        Table(name="googl", zone="ROUGE")


def atest_sql_schema():
    types = Table.model_serialized_types()
    print(types)
    assert types == {
        "name": "string",
        "columns": [
            {
                "name": "string",
                "type": "string",
                "comment": "string",
                "catalog_name": "string",
                "schema_name": "string",
                "table_name": "string",
                "unit": "string",
                "pii": "boolean",
                "func_name": "string",
                "input_cols": ["string"],
                "func_kwargs": {},
                "jsonize": "boolean",
            }
        ],
        "primary_key": "string",
        "comment": "string",
        "catalog_name": "string",
        "schema_name": "string",
        "grants": [{"principal": "string", "privileges": ["string"]}],
        "data": [[None]],
        "timestamp_key": "string",
        "event_source": {
            "name": "string",
            "description": "string",
            "producer": {"name": "string", "description": "string", "party": "integer"},
            "events_root": "string",
            "read_as_stream": "boolean",
            "type": "string",
            "fmt": "string",
            "multiline": "boolean",
        },
        "table_source": {
            "read_as_stream": "boolean",
            "name": "string",
            "schema_name": "string",
            "catalog_name": "string",
        },
        "zone": "string",
        "pipeline_name": "string",
    }

    types = Column.model_serialized_types()
    assert types == {
        "name": "string",
        "type": "string",
        "comment": "string",
        "catalog_name": "string",
        "schema_name": "string",
        "table_name": "string",
        "unit": "string",
        "pii": "boolean",
        "func_name": "string",
        "input_cols": ["string"],
        "func_kwargs": {},
        "jsonize": "boolean",
    }

    types["func_kwargs"] = "string"
    schema = Column.model_sql_schema(types)
    assert (
        schema
        == "(name string, type string, comment string, catalog_name string, schema_name string, table_name string, unit string, pii boolean, func_name string, input_cols ARRAY<string>, func_kwargs string, jsonize boolean)"
    )


def atest_create_and_insert():
    # Timestamp is included in catalog name to prevent conflicts when running
    # multiple tests in parallel
    cat_name = "laktory_testing_" + str(datetime.now().timestamp()).replace(".", "")

    cat = Catalog(
        name=cat_name,
    )
    cat.create(if_not_exists=True)
    db = Schema(name="default", catalog_name=cat_name)
    db.create()
    table = Table(
        catalog_name=cat_name,
        schema_name="default",
        name="stocks",
        columns=[
            {
                "name": "open",
                "type": "double",
            },
            {
                "name": "close",
                "type": "double",
            },
            {
                "name": "symbol",
                "type": "string",
            },
            {
                "name": "d",
                "type": "string",
                "jsonize": True,
            },
        ],
        data=[
            [1, 2, "googl", {"a": 1}],
            [3, 4, "googl", {"b": 2}],
            [5, 6, "googl", {"c": 3}],
        ],
    )
    table.create(or_replace=True, insert_data=True)
    assert table.exists()
    data = table.select(load_json=True)
    assert data == [
        ["1.0", "2.0", "googl", {"a": 1}],
        ["3.0", "4.0", "googl", {"b": 2}],
        ["5.0", "6.0", "googl", {"c": 3}],
    ]
    table.delete()
    cat.delete(force=True)


def atest_meta():
    # TODO: Remove - deprecated
    meta = table.meta_table()

    meta.catalog_name = "main"

    assert "catalog_name" in meta.column_names
    assert "schema_name" in meta.column_names
    assert "name" in meta.column_names
    assert "comment" in meta.column_names
    assert "columns" in meta.column_names

    is_found = False
    for c in meta.columns:
        if c.name == "event_source":
            is_found = True
            assert (
                c.type
                == "STRUCT<name: string, description: string, producer: STRUCT<name: string, description: string, party: integer>, events_root: string, read_as_stream: boolean, type: string, fmt: string, multiline: boolean>"
            )
    assert is_found

    dump = meta.model_dump()
    col_dump = dump["columns"][0]
    assert list(col_dump.keys()) == [
        "name",
        "type",
        "comment",
        "catalog_name",
        "schema_name",
        "table_name",
        "unit",
        "pii",
        "func_name",
        "input_cols",
        "func_kwargs",
        "jsonize",
    ]


if __name__ == "__main__":
    test_model()
    test_data()
    # atest_sql_schema()
    # atest_create_and_insert()
    # atest_meta()



import pytest

from laktory.models import DataEvent
from laktory.models import DataEventHeader
from datetime import datetime
from zoneinfo import ZoneInfo

event = DataEvent(
    name="flight_record",
    producer={
        "name": "FDR",
    },
    data={
        "created_at": datetime(2023, 7, 1, 1, 0, 0),
        "airspeed": 100.0,
        "altitude": 20000.0,
    },
)


def test_dataevent_header():
    header = DataEventHeader(
        name="flight_record",
        description=None,
        producer={
            "name": "FDR",
            "party": 1,
        },
    )
    assert header.model_dump() == {
        "name": "flight_record",
        "description": None,
        "producer": {"name": "FDR", "description": None, "party": 1},
        "events_root": "/Volumes/dev/sources/landing/events/",
    }
    assert header.event_root == "/Volumes/dev/sources/landing/events/FDR/flight_record/"


def test_dataevent():
    assert event.producer.name == "FDR"
    assert event.data["altitude"] == 20000
    assert event.data["_producer_name"] == "FDR"
    assert event.data["_created_at"] == datetime(
        2023, 7, 1, 1, 0, 0, tzinfo=ZoneInfo("UTC")
    )
    assert event.created_at == datetime(2023, 7, 1, 1, 0, 0, tzinfo=ZoneInfo("UTC"))
    assert (
        event.get_landing_filepath()
        == "/Volumes/dev/sources/landing/events/FDR/flight_record/2023/07/01/flight_record_20230701T010000000Z.json"
    )


def test_model_dump():
    # Without exclusions
    d = event.model_dump(exclude=[])
    print(d)
    for k, v in d.items():
        print(k, v)
    assert d == {
        "event_name": "flight_record",
        "event_description": None,
        "event_producer": {"name": "FDR", "description": None, "party": 1},
        "events_root": "/Volumes/dev/sources/landing/events/",
        "data": {
            "created_at": "2023-07-01T01:00:00",
            "airspeed": 100.0,
            "altitude": 20000.0,
            "_name": "flight_record",
            "_producer_name": "FDR",
            "_created_at": "2023-07-01T01:00:00Z",
        },
        "tstamp_col": "created_at",
    }

    # With exclusions
    d = event.model_dump()
    print(d.keys())
    assert list(d.keys()) == [
        "event_name",
        "event_description",
        "event_producer",
        "data",
    ]
    # for k, v in d.items():
    #     print(k, v)


def test_to_azure_storage_container():
    try:
        import azure.storage
    except ModuleNotFoundError:
        return

    event.to_azure_storage_container(container_name="unit-testing", overwrite=True)
    with pytest.raises(FileExistsError):
        event.to_azure_storage_container(container_name="unit-testing")
    event.to_azure_storage_container(container_name="unit-testing", skip_if_exists=True)


def test_to_aws_s3_bucket():
    try:
        import boto3
    except ModuleNotFoundError:
        return

    event.to_aws_s3_bucket(bucket_name="okube-unit-testing", overwrite=True)
    with pytest.raises(FileExistsError):
        event.to_aws_s3_bucket(bucket_name="okube-unit-testing")
    event.to_aws_s3_bucket(bucket_name="okube-unit-testing", skip_if_exists=True)


def test_to_databricks_mount():
    # TODO: add tests
    pass
    # event.to_databricks_mount()


if __name__ == "__main__":
    test_dataevent_header()
    test_dataevent()
    test_model_dump()
    test_to_azure_storage_container()
    test_to_aws_s3_bucket()
    test_to_databricks_mount()



import time
from databricks.sdk import WorkspaceClient as _WorkspaceClient
from databricks.sdk.service.sql import StatementState

from laktory import settings


class WorkspaceClient(_WorkspaceClient):
    def __init__(self, *args, **kwargs):
        kwargs["host"] = kwargs.get("host", settings.databricks_host)
        kwargs["token"] = kwargs.get("token", settings.databricks_token)
        super().__init__(*args, **kwargs)

    def execute_statement_and_wait(
        self, statement, warehouse_id=None, catalog_name=None
    ):
        # Settings
        if warehouse_id is None:
            warehouse_id = settings.databricks_warehouse_id

        r = self.statement_execution.execute_statement(
            statement=statement,
            catalog=catalog_name,
            warehouse_id=warehouse_id,
        )
        statement_id = r.statement_id
        state = r.status.state

        while state in [StatementState.PENDING, StatementState.RUNNING]:
            r = self.statement_execution.get_statement(statement_id)
            time.sleep(1)
            state = r.status.state

        if state != StatementState.SUCCEEDED:
            # TODO: Create specific error
            print(r.status.error.message)
            raise Exception(r.status.error)

        return r



from laktory.spark import DataFrame
from typing import Union

from laktory.models.datasources.basedatasource import BaseDataSource
from laktory._logger import get_logger

logger = get_logger(__name__)


class TableDataSource(BaseDataSource):
    name: Union[str, None]
    schema_name: Union[str, None] = None
    catalog_name: Union[str, None] = None

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.name
        else:
            name += f".{self.name}"

        return name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def read(self, spark) -> DataFrame:
        from laktory.dlt import read
        from laktory.dlt import read_stream

        if self.read_as_stream:
            logger.info(f"Reading {self.full_name} as stream")
            df = read_stream(self.full_name)
        else:
            logger.info(f"Reading {self.full_name} as static")
            df = read(self.full_name)

        return df



import os
from datetime import datetime

from laktory.models import Catalog
from laktory.models import Table
from laktory.models import Pipeline
from laktory._testing import StockPricesPipeline

root_dir = os.path.dirname(__file__)


def test_pipeline():
    pl = StockPricesPipeline()

    assert pl.tables[0].zone == "BRONZE"
    assert pl.model_dump()["tables"][0]["zone"] == "BRONZE"


def test_read_yaml():
    with open(f"{root_dir}/pl-stocks.yaml", "r") as fp:
        pl = Pipeline.model_validate_yaml(fp)

    assert pl.name == "pl-stocks"


def atest_tables_meta():
    # TODO: Remove - Deprecated
    pl = StockPricesPipeline()

    table = pl.get_tables_meta()
    df = table.df
    assert df["name"].tolist() == ["brz_stock_prices", "slv_stock_prices"]
    assert df["zone"].tolist() == ["BRONZE", "SILVER"]
    assert df["pipeline_name"].tolist() == ["pl-stock-prices", "pl-stock-prices"]
    assert df["comment"].tolist() == [None, None]


def atest_columns_meta():
    pl = StockPricesPipeline()

    table = pl.get_columns_meta()
    df = table.df
    assert df["name"].tolist() == ["created_at", "low", "high"]
    assert df["type"].tolist() == ["timestamp", "double", "double"]


def atest_publish_meta():
    # TODO: Remove - deprecated
    # Timestamp is included in catalog name to prevent conflicts when running
    # multiple tests in parallel
    catalog_name = "laktory_testing_" + str(datetime.now().timestamp()).replace(".", "")

    pl = StockPricesPipeline()
    pl.publish_tables_meta(catalog_name=catalog_name)

    tables = Table(name="tables", schema_name="laktory", catalog_name=catalog_name)
    data = tables.select()
    assert data[0][0] == "brz_stock_prices"

    columns = Table(name="columns", schema_name="laktory", catalog_name=catalog_name)
    data = columns.select()
    assert data[1][0] == "low"

    # Cleanup
    Catalog(name=catalog_name).delete(force=True)


if __name__ == "__main__":
    test_pipeline()
    test_read_yaml()
    # atest_tables_meta()
    # atest_columns_meta()
    # atest_publish_meta()



from datetime import datetime

from laktory.models import InitScript


def test_init_script():
    init_script = InitScript(
        source="../libraries/init_scripts/install_laktory.sh",
        permissions=[{"permission_level": "CAN_READ", "group_name": "account users"}],
    )
    assert init_script.filename == "install_laktory.sh"
    assert init_script.key == "init_scripts-install_laktory"
    assert init_script.path == "/init_scripts/install_laktory.sh"

    assert init_script.permissions[0].permission_level == "CAN_READ"
    assert init_script.permissions[0].group_name == "account users"


if __name__ == "__main__":
    test_init_script()



from typing import Literal
from laktory.models.base import BaseModel

PRIVILEGES = [
    "ALL_PRIVILEGES",
    "SELECT",
    "MODIFY",
]


class TableGrant(BaseModel):
    principal: str
    privileges: list[Literal[tuple(PRIVILEGES)]]


t = TableGrant(principal="a", privileges=["SELECT"])



from typing import Union
from pydantic import computed_field
from pydantic import field_validator

from laktory.contants import SUPPORTED_TYPES
from laktory.models.base import BaseModel


class Column(BaseModel):
    name: str
    type: str = "string"
    comment: Union[str, None] = None
    catalog_name: Union[str, None] = None
    schema_name: Union[str, None] = None
    table_name: Union[str, None] = None
    unit: Union[str, None] = None
    pii: Union[bool, None] = None
    func_name: Union[str, None] = None
    input_cols: list[str] = []
    func_kwargs: dict[str, Union[str, None]] = {}
    jsonize: bool = False

    @field_validator("type")
    def default_load_path(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_TYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_TYPES}"
                )
        return v

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        if self.table_name:
            if _id == "":
                _id = self.table_name
            else:
                _id += f".{self.table_name}"

        return _id

    @property
    def full_name(self) -> str:
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def database_name(self) -> str:
        return self.schema_name

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    # TODO: Move to Databricks SDK engine
    # @classmethod
    # def meta_table(cls):
    #     from laktory.models.sql.table import Table
    #
    #     # Build columns
    #     columns = []
    #     for k, t in cls.model_serialized_types().items():
    #         jsonize = False
    #         if k in ["func_kwargs"]:
    #             t = "string"
    #             jsonize = True
    #
    #         columns += [
    #             Column(name=k, type=py_to_sql(t, mode="schema"), jsonize=jsonize)
    #         ]
    #
    #     # Set table
    #     return Table(
    #         name="columns",
    #         schema_name="laktory",
    #         columns=columns,
    #     )



import os

from laktory.models import Column

GOOGL = {
    "name": "close",
    "type": "double",
    "unit": "USD",
    "catalog_name": "lakehouse",
    "schema_name": "markets",
    "table_name": "googl",
}

root_dir = os.path.dirname(__file__)


def test_model():
    c0 = Column(**GOOGL)
    c1 = Column.model_validate(GOOGL)
    assert c1.type == "double"
    assert c1.catalog_name == "lakehouse"
    assert c1.schema_name == "markets"
    assert c1.table_name == "googl"
    assert c1.full_name == "lakehouse.markets.googl.close"
    assert "func_name" in c1.model_fields
    assert c0 == c1


def test_read():
    c0 = Column(**GOOGL)

    with open(f"{root_dir}/googl.yaml", "r") as fp:
        c1 = Column.model_validate_yaml(fp)

    with open(f"{root_dir}/googl.json", "r") as fp:
        c2 = Column.model_validate_json_file(fp)

    assert c1 == c0
    assert c2 == c0


if __name__ == "__main__":
    test_model()
    test_read()



from laktory.models.compute.initscript import InitScript
from laktory.models.compute.cluster import Cluster
from laktory.models.compute.warehouse import Warehouse
from laktory.models.compute.notebook import Notebook
from laktory.models.compute.pipeline import Pipeline
from laktory.models.compute.job import Job
